#steganography
# Python script which will modify the R,G,B values of each pixel in a specified
# PNG image in order to hide a text file. The text file is revealed by subtracting
# the original image from the encoded image. Each text character is converted into an
# 8-bit ASCII code (i.e. an integer 000 to 255). The first digit will modify the Red
# colour, the second the Blue colour and the third the Green. In order for text data to
# be added to any image, the maximum colour values are 255-9 = 246.
#
# LAST UPDATED by Andy French Jan 2018

#Import timing, GUI, numeric and plotting libraries
import time
import tkinter as tk
import numpy as np
from tkinter import filedialog
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.image as mpimg

#

#Image import and RGB segmentation function. Call as:
# R, G, B, rows, cols = rgb_import( image_filename )
def rgb_import( image_filename ):

    #Import PNG image and extract R,G,B colour planes. These will be
    #stored as numpy arrays of dimensions rows x cols, with floating point
    #numbert elements betwen 0 and 1
    print(" \n Importing PNG image " + image_filename + " ... ")
    rgb= mpimg.imread( image_filename )
    dims = np.shape(rgb)
    rows = dims[0]
    cols = dims[1]
    R = np.zeros((rows,cols))
    G = np.zeros((rows,cols))
    B = np.zeros((rows,cols))
    if len(dims)==2:
        #Grayscale
        R[:,:] = (1/3)*rgb[:,:]
        G[:,:] = (1/3)*rgb[:,:]
        B[:,:] = (1/3)*rgb[:,:]
        isgray = 1
    else:
        #Full colour
        R[:,:] = rgb[:,:,0]
        G[:,:] = rgb[:,:,1]
        B[:,:] = rgb[:,:,2]
        isgray = 0

    #Remove unwanted variables from memory
    del dims, rgb

    #Set return values of function
    return R, G, B, rows, cols

#

#Function to return a row vector from a numpy n.d array a
def flatten(a):
    b = np.reshape( a, (1,np.size(a)) )
    b = b[0]
    return b

#

#Display greeting message to console
print("\n STEGANOGRAPHY")
print(" Welcome to the Python Steganography machine! ")
print(" By Dr Andrew French. Jan 2018 " )

#

#Choose and import 'original' PNG image
root = tk.Tk()
root.withdraw()
image_file = filedialog.askopenfilename(title = "Choose original PNG image...",filetypes=[("PNG images","*.png")])

#Import image and perform colour separations
Ro, Go, Bo, rows, cols = rgb_import( image_file )

#Convert R,G,B arrays such that max value is 255 (i.e. same as 8 bit integers)
Ro = 255*Ro
Go = 255*Go
Bo = 255*Bo

#Limit max values to 246 to enable 9 (the max ASCII digit) to be added
Ro[ Ro > 246 ] = 246
Go[ Go > 246 ] = 246
Bo[ Bo > 246 ] = 246

#

#Request encryption/decryption mode
encryption_mode = input("\n Select Encrypt (press letter e) or Decrypt mode (anything else) ... \n > ")

# ENCRYPTION MODE
# - Choose text file to hide in image
# - Modify RGB values of original image to hide text file
# - Saved encoded PNG image
if encryption_mode=="e":

    #Choose and import text file to be hidden in orginal image
    root = tk.Tk()
    root.withdraw()
    text_file = filedialog.askopenfilename(title = "Choose text file to hide...",filetypes=[("Text files","*.txt")])

    #Store time since current epoch (1970?) in s, and hence start an internal stopwatch.
    t0 = time.time()
    t1 = t0

    #Load text file and extract string information
    print("\n Loading text file: " + text_file )
    f = open( text_file,'r' )
    s = f.read()
    N = len(s)

    #Step through s and find digits associated with the 8-bit ASCII code
    print(" \n Extracting ASCII codes and modifying R,G,B values of original image ... " )
    hundreds = np.zeros(N)
    tens = np.zeros(N)
    ones = np.zeros(N)
    ii = 0
    for i in range(0,N):
        #Find three-digit ASCII code e.g. "007" or "042" or "123"
        digits = str( ord( s[i] ) ).zfill(3)
        hundreds[i] = float( digits[0] )
        tens[i] = float( digits[1] )
        ones[i] = float( digits[2] )
        ii = ii + 1
        if ii > 200000:
            print(" *",end="",flush=True)
            ii = 0

    #Truncate text file if longer than number of pixels in original image              
    if N > Ro.size:
        hundreds = hundreds[0:Ro.size]
        tens = tens[0:Ro.size]
        ones = ones[0:Ro.size]
        N = Ro.size
    else:
        #Pack with zeros to match size of R,G,B arrays
        hundreds = np.append( hundreds, np.zeros( Ro.size - N ) )
        tens = np.append( tens, np.zeros( Ro.size - N ) )
        ones = np.append( ones, np.zeros( Ro.size - N ) )

    #Reshape ASCII numbers to match R,G,B dimensions
    hundreds = np.reshape( hundreds, (rows,cols) )
    tens = np.reshape( tens, (rows,cols) )
    ones = np.reshape( ones, (rows,cols) )

    #Add ASCII data to R,G,B values of orginal image
    R = Ro + hundreds
    G = Go + tens
    B = Bo + ones

    #Clear unwanted variables from memory
    del Ro, Go, Bo, hundreds, tens, ones, s

    #Assemble RGB array for encoded image 
    rgbe = np.zeros((rows,cols,3))
    rgbe[:,:,0] = (1/255)*R[:,:]
    rgbe[:,:,1] = (1/255)*G[:,:]
    rgbe[:,:,2] = (1/255)*B[:,:]

    #Clear unwanted variables from memory
    del R, G, B

    #Save encoded PNG image
    print(" \n Saving encoded PNG file ... " )
    plt.imsave( image_file.replace(".png","_encoded.png"), rgbe )

#

# DECRYPTION MODE
# - Choose and import encoded image
# - Subtract original R,G,B values from encoded image
# - Extract ASCII codes from remainder and hence construct text file
# - Write out text file
else:

    #Choose and import encoded PNG image
    root = tk.Tk()
    root.withdraw()
    image_file = filedialog.askopenfilename(title = "Choose encoded PNG image...",filetypes=[("PNG images","*.png")])

    #Store time since current epoch (1970?) in s, and hence start an internal stopwatch.
    t0 = time.time()
    t1 = t0

    #Import image and perform colour separations
    R, G, B, rows, cols = rgb_import( image_file )

    #Convert R,G,B arrays such that max value is 255 (i.e. same as 8 bit integers)
    #and subtract original image values
    dR = flatten( 255*R - Ro )
    dG = flatten( 255*G - Go )
    dB = flatten( 255*B - Bo )

    #Clear unwanted variables from memory
    del R, G, B, Ro, Go, Bo

    #Step through residuals above, extract ASCII code and construct string s
    s = ""
    i = 0
    ii = 0
    a = 1
    print(" \n Extracting ASCII codes from encoded and orginal PNG file differences ... " )
    while (a > 0) & ( i < np.size(dR) ):
        a = int( np.around( 100*dR[i] + 10*dG[i] + dB[i] ) )
        s = s + chr(a)
        i = i + 1
        ii = ii + 1
        if ii > 200000:
            print(" *",end="",flush=True)
            ii = 0
            
    #Write string to a text file
    print(" \n Saving text file ... " )
    f = open( image_file.replace(".png","_hidden_text.txt"), "w" )
    f.write(s)
    f.close()

#Wait for user input to close terminal window
print("\n Total processing time is: " + str(time.time() - t0) + " seconds.")
input("\n Press return to close ")

#End of code
