#cryptosonic
# Python script which will hide a text file in a sound file by modifying the sample values by the
# ASCII codes for each character in the text file. One ASCII character will modify one sound sample,
# in alternate left and right channels, correspondiong to the ASCII code 8-bit integer.
#
# LAST UPDATED by Andy French Jan 2018

#Import timing, GUI, numeric and sound libraries
import tkinter as tk
import numpy as np
from tkinter import filedialog
import wave, struct, time

#

#Read sound file function. Returns left and right channel samples (16 bit integers)
#number of samples (ns), sample rate fs (Hz) and time vector (s)
def wavread( wav_filename ):

    #Open .wav file
    w = wave.open( wav_filename, 'r' )

    #Extract sample rate /Hz
    fs = w.getframerate()
    ns = w.getnframes()

    #Extract 16-bit integer numpy arrays for left and right
    #stereo channels. If original file was mono, replicate
    da = np.fromstring( w.readframes(ns), dtype=np.int16 )
    w.close()
    if w.getnchannels()==1:
        left, right = da[:], da[:]
    else:
        left, right = da[0::2], da[1::2]

    #Generate time vector
    ns = len(left)
    t = np.linspace(0, (ns-1)/fs, ns)

    #Return function outputs
    return left, right, ns, fs, t

#

#Sound export function. Adapted from: http://blog.acipo.com/wave-generation-in-python/
def wavwrite( wav_filename, ns, fs, left, right  ):

    #Open .wav file
    wf = wave.open( wav_filename,'w')

    #Set parameters of wf object
    wf.setnchannels(2) # stereo
    wf.setsampwidth(2) # assume 16-bit audio (and int16 left and right samples)
    wf.setframerate(fs)
    wf.setnframes(ns)

    #Write sample data
    ii = 0
    for i in range(0,ns):
        l = int(left[i])
        r = int(right[i])
        wf.writeframesraw( struct.pack('<hh', l, r ) )
        ii = ii + 1
        if ii > fs:
            print(" *",end="",flush=True)
            ii = 0
    wf.close()

#

#Display greeting message to console
print("\n CRYPTOSONIC")
print(" Welcome to the Python Sonic-Steganography machine! ")
print(" By Dr Andrew French. Jan 2018 " )

#

#Choose and import 'original' sound image
root = tk.Tk()
root.withdraw()
wav_file = filedialog.askopenfilename(title = "Choose original .wav file...",filetypes=[("WAV files","*.wav")])

#Import original sound file
left, right, ns, fs, t = wavread( wav_file )

#If needed, reduce sound data to ensure ASCII code digits up to 255 can be added to it.
#Assume sound file is 15 bits + one bit of sign (i.e. max amplitude is 2**15-1)
max_s = 2**15 - 1 - 255
right[ right > max_s ] = max_s
left[ left > max_s ] = max_s

#

#Request encryption/decryption mode
encryption_mode = input("\n Select Encrypt (press letter e) or Decrypt mode (anything else) ... \n > ")

# ENCRYPTION MODE
if encryption_mode=="e":

    #Choose and import text file to be hidden in orginal sound file
    root = tk.Tk()
    root.withdraw()
    text_file = filedialog.askopenfilename(title = "Choose text file to hide...",filetypes=[("Text files","*.txt")])

    #Store time since current epoch (1970?) in s, and hence start an internal stopwatch.
    t0 = time.time()
    t1 = t0

    #Load text file and extract string information
    print("\n Loading text file: " + text_file )
    f = open( text_file,'r' )
    txt = f.read()
    N = len(txt)

    #Check if N is even or odd
    if N%2==0:
        #Even
        imax = N
    else:
        #Odd
        imax = N-1

    #Step through s and find digits associated with the 8-bit ASCII code. Add these to the sound file values
    print(" \n Extracting ASCII codes and modifying sample values of original sound file ... " )
    i = 0
    j = 0
    ii = 0
    right_e = right.copy()
    left_e = left.copy()
    while (i < imax ) & (j<ns):

        #Add ASCII code integer to sound samples, alternately left and right channels
        left_e[j] = left[j] + ord( txt[i] )
        right_e[j] = right[j] + ord( txt[i+1] )
        i = i + 2
        j = j + 1
        ii = ii + 2
        if ii > 200000:
            print(" *",end="",flush=True)
            ii = 0

    #Check if one character remaining
    if N%2==1:
        left_e[j] = left[j] + ord( txt[N-1] )
        
    #Save encoded sound file
    print(" \n Saving encoded .wav file ... " )
    wavwrite( wav_file.replace(".wav","_encoded.wav"), ns, fs, left_e, right_e  )

#

# DECRYPTION MODE
else:

    #Choose encoded sound file
    root = tk.Tk()
    root.withdraw()
    wav_file = filedialog.askopenfilename(title = "Choose encoded .wav file...",filetypes=[("WAV files","*.wav")])

    #Store time since current epoch (1970?) in s, and hence start an internal stopwatch.
    t0 = time.time()
    t1 = t0

    #Import encoded sound file and substract original
    left_e, right_e, ns, fs, t = wavread( wav_file )
    dleft = left_e - left
    dright = right_e - right
    del left_e, right_e, left, right

    #Step through residual above, extract ASCII code and construct string s
    txt = [""]*( int(ns/2) + 1 )
    i = 0
    j = 0
    ii = 0
    a = 1
    print(" \n Extracting ASCII codes from encoded and orginal sound file differences ... " )
    while ( a > 0) & ( j < ns ):
        txt[i] = chr( dleft[j] )
        txt[i+1] = chr( dright[j] )
        a = dleft[j] + dright[j]
        i = i + 2
        j = j + 1
        ii = ii + 2
        if ii > 200000:
            print(" *",end="",flush=True)
            ii = 0

    #Remove any unused spaces from txt
    txt = txt[0:i]
            
    #Write string to a text file
    print(" \n Saving text file ... " )
    f = open( wav_file.replace(".wav","_hidden_text.txt"), "w" )
    f.write("".join(txt))
    f.close()

#Wait for user input to close terminal window
print("\n Total processing time is: " + str(time.time() - t0) + " seconds.")
input("\n Press return to close ")

#End of code
